from .restart_section import RestartSection
from .distinct_section import DistinctSection
from .connection_section import ConnectionSection
from .query_section import QuerySection
from .rs_state_section import RsStateSection
from .rs_info_section import RsInfoSection



#!/bin/python
"""Command line tool utility."""

import argparse
import codecs
import datetime
import os
import re
import signal
import sys

from dateutil.tz import tzutc

from mtools.util.logfile import LogFile
from mtools.version import __version__

try:
    from mtools.util.profile_collection import ProfileCollection

    class InputSourceAction(argparse.FileType):
        """
        Extend the FileType class from the argparse module.

        Try to open the file and pass the handle to a new LogFile object, but
        if that's not possible it will catch the exception and interpret the
        string as a MongoDB URI and try to connect to the database. In that
        case, it will return a ProfileCollection object.

        Both derive from the same base class InputSource and support iteration
        over LogEvents.
        """

        def __call__(self, string):
            """Open log file or MongoDB database."""

            try:
                # catch filetype and return LogFile object
                filehandle = argparse.FileType.__call__(self, string)
                return LogFile(filehandle)

            except argparse.ArgumentTypeError:
                # not a file, try open as MongoDB database
                m = re.match('^(\w+)(?::(\d+))?(?:/([a-zA-Z0-9._-]+))?$',
                             string)
                if m:
                    hostname, port, namespace = m.groups()
                    port = int(port) if port else 27017
                    namespace = namespace or 'test.system.profile'
                    if '.' in namespace:
                        database, collection = namespace.split('.', 1)
                    else:
                        database = namespace
                        collection = 'system.profile'

                    if (hostname == 'localhost' or
                            re.match('\d+\.\d+\.\d+\.\d+', hostname)):
                        return ProfileCollection(hostname, port, database,
                                                 collection)

                raise argparse.ArgumentTypeError("can't open %s as file or "
                                                 "MongoDB connection string."
                                                 % string)

except ImportError:
    class InputSourceAction(argparse.FileType):
        """Extend the FileType class from the argparse module."""

        def __call__(self, string):
            """Open log file."""
            try:
                # catch filetype and return LogFile object
                filehandle = argparse.FileType.__call__(self, string)
                return LogFile(filehandle)
            except argparse.ArgumentTypeError:
                raise argparse.ArgumentTypeError("can't open %s" % string)


class BaseCmdLineTool(object):
    """
    Base class for any mtools command line tool.

    Adds --version flag and basic control flow.
    """

    def __init__(self):
        """
        Constructor.

        Any inheriting class should add a description to the argparser and
        extend it with additional arguments as needed.
        """
        # define argument parser and add version argument
        self.argparser = argparse.ArgumentParser()
        self.argparser.add_argument('--version', action='version',
                                    version="mtools version {0} || Python {1}".format(
                                        __version__, sys.version))
        self.argparser.add_argument('--no-progressbar', action='store_true',
                                    default=False,
                                    help='disables progress bar')
        self.is_stdin = not sys.stdin.isatty()

        # Set stdout encoding to utf-8 if not set
        # Need to check for nose because it monkey patches sys.stdout:
        #     https://github.com/nose-devs/nose/issues/1065
        if ('nose' not in sys.modules.keys()
                and hasattr(sys.stdout, 'encoding')
                and not sys.stdout.encoding):
            sys.stdout = codecs.getwriter('utf-8')(sys.stdout)

    def run(self, arguments=None, get_unknowns=False):
        """
        Init point to execute the script.

        If `arguments` string is given, will evaluate the arguments, else
        evaluates sys.argv. Any inheriting class should extend the run method
        (but first calling BaseCmdLineTool.run(self)).
        """

        # redirect PIPE signal to quiet kill script, if not on Windows
        if os.name != 'nt':
            signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        if get_unknowns:
            if arguments:
                self.args, self.unknown_args = (self.argparser.parse_known_args
                                                (args=arguments.split()))
            else:
                (self.args,
                 self.unknown_args) = self.argparser.parse_known_args()
            self.args = vars(self.args)
        else:
            if arguments:
                myargs = arguments.split()
                self.args = vars(self.argparser.parse_args
                                 (args=myargs))
            else:
                self.args = vars(self.argparser.parse_args())

        self.progress_bar_enabled = (not (self.args['no_progressbar'] or
                                          self.is_stdin))

    def _datetime_to_epoch(self, dt):
        """Convert the datetime to unix epoch (properly)."""
        if dt:
            td = (dt - datetime.datetime.fromtimestamp(0, tzutc()))
            # don't use total_seconds(), that's only available in 2.7
            total_secs = int((td.microseconds +
                              (td.seconds + td.days * 24 * 3600) *
                              10**6) / 10**6)
            return total_secs
        else:
            return 0

    def update_progress(self, progress, prefix=''):
        """
        Print a progress bar for longer-running scripts.

        The progress value is a value between 0.0 and 1.0. If a prefix is
        present, it will be printed before the progress bar.
        """
        total_length = 40

        if progress == 1.:
            sys.stderr.write('\r' + ' ' * (total_length + len(prefix) + 50))
            sys.stderr.write('\n')
            sys.stderr.flush()
        else:
            bar_length = int(round(total_length * progress))
            sys.stderr.write('\r%s [%s%s] %.1f %% '
                             % (prefix, '=' * bar_length,
                                ' ' * (total_length - bar_length),
                                progress * 100))
            sys.stderr.flush()


class LogFileTool(BaseCmdLineTool):
    """Base class for any mtools tool that acts on logfile(s)."""

    def __init__(self, multiple_logfiles=False, stdin_allowed=True):
        """Add logfile(s) and stdin option to the argument parser."""
        BaseCmdLineTool.__init__(self)

        self.multiple_logfiles = multiple_logfiles
        self.stdin_allowed = stdin_allowed

        arg_opts = {'action': 'store', 'type': InputSourceAction('rb')}

        if self.multiple_logfiles:
            arg_opts['nargs'] = '*'
            arg_opts['help'] = 'logfile(s) to parse'
        else:
            arg_opts['help'] = 'logfile to parse'

        if self.is_stdin:
            if not self.stdin_allowed:
                raise SystemExit("this tool can't parse input from stdin.")

            arg_opts['const'] = LogFile(sys.stdin)
            arg_opts['action'] = 'store_const'
            if 'type' in arg_opts:
                del arg_opts['type']
            if 'nargs' in arg_opts:
                del arg_opts['nargs']
        self.argparser.add_argument('logfile', **arg_opts)


if __name__ == '__main__':
    tool = LogFileTool(multiple_logfiles=True, stdin_allowed=True)
    tool.run()
    print(tool.args)
    # for line in tool.args['logfile']:
    #     print(line)



#!/bin/python

import json
import re

import six


def _decode_pattern_list(data):
    rv = []
    contains_dict = False
    for item in data:
        if isinstance(item, six.text_type):
            item = item.encode('utf-8')
        elif isinstance(item, list):
            item = _decode_pattern_list(item)
        elif isinstance(item, dict):
            item = _decode_pattern_dict(item)
            contains_dict = True
        rv.append(item)

    # avoid sorting if any element in the list is a dict
    if not contains_dict:
        rv = sorted(rv)

    return rv


def _decode_pattern_dict(data):
    rv = {}
    for key, value in six.iteritems(data):
        if isinstance(key, bytes):
            key = key.encode('utf-8')
        if isinstance(key, six.text_type):
            if key in ['$in', '$gt', '$gte', '$lt', '$lte', '$exists']:
                return 1
            if key == '$nin':
                value = 1
            if key in ['query', '$query']:
                return _decode_pattern_dict(value)

        if isinstance(value, list):
            value = _decode_pattern_list(value)
        elif isinstance(value, dict):
            value = _decode_pattern_dict(value)
        else:
            value = 1
        rv[key] = value
    return rv


def shell2json(s):
    """Convert shell syntax to json."""
    replace = {
        r'BinData\(.+?\)': '1',
        r'(new )?Date\(.+?\)': '1',
        r'Timestamp\(.+?\)': '1',
        r'ObjectId\(.+?\)': '1',
        r'DBRef\(.+?\)': '1',
        r'undefined': '1',
        r'MinKey': '1',
        r'MaxKey': '1',
        r'NumberLong\(.+?\)': '1',
        r'/.+?/\w*': '1'
    }

    for key, value in replace.items():
        s = re.sub(key, value, s)

    return s


def json2pattern(s):
    """
    Convert JSON format to a query pattern.

    Includes even mongo shell notation without quoted key names.
    """
    # make valid JSON by wrapping field names in quotes
    s, _ = re.subn(r'([{,])\s*([^,{\s\'"]+)\s*:', ' \\1 "\\2" : ', s)
    # handle shell values that are not valid JSON
    s = shell2json(s)
    # convert to 1 where possible, to get rid of things like new Date(...)
    s, n = re.subn(r'([:,\[])\s*([^{}\[\]"]+?)\s*([,}\]])', '\\1 1 \\3', s)
    # now convert to dictionary, converting unicode to ascii
    try:
        doc = json.loads(s, object_hook=_decode_pattern_dict)
        return json.dumps(doc, sort_keys=True, separators=(', ', ': '))
    except ValueError as ex:
        return None


if __name__ == '__main__':

    s = ('{d: {$gt: 2, $lt: 4}, b: {$gte: 3}, '
         'c: {$nin: [1, "foo", "bar"]}, "$or": [{a:1}, {b:1}] }')
    print(json2pattern(s))

    s = ('{a: {$gt: 2, $lt: 4}, '
         '"b": {$nin: [1, 2, 3]}, "$or": [{a:1}, {b:1}] }')
    print(json2pattern(s))

    s = ('{a: {$gt: 2, $lt: 4}, '
         '"b": {$nin: [1, 2, 3]}, "$or": [{a:1}, {b:1}] }')
    print(json2pattern(s))

    s = ("{a: {$gt: 2, $lt: 4}, "
         "b: {$in: [ ObjectId('1234564863acd10e5cbf5f6e'), "
         "ObjectId('1234564863acd10e5cbf5f7e') ] } }")
    print(json2pattern(s))

    s = ("{ sk: -1182239108, "
         "_id: { $in: [ ObjectId('1234564863acd10e5cbf5f6e'), "
         "ObjectId('1234564863acd10e5cbf5f7e') ] } }")
    print(json2pattern(s))

    s = '{ a: 1, b: { c: 2, d: "text" }, e: "more test" }'
    print(json2pattern(s))

    s = ('{ _id: ObjectId(\'528556616dde23324f233168\'), '
         'config: { _id: 2, host: "localhost:27017" }, ns: "local.oplog.rs" }')
    print(json2pattern(s))



#!/usr/bin/env python

import os
import sys
import webbrowser

import mtools
from mtools.util.cmdlinetool import LogFileTool


class MLogVisTool(LogFileTool):

    def __init__(self):
        LogFileTool.__init__(self, multiple_logfiles=False, stdin_allowed=True)

        self.argparser.description = ('mongod/mongos log file visualizer '
                                      '(browser edition). Extracts '
                                      'information from each line of the log '
                                      'file and outputs a html file that can '
                                      'be viewed in a browser. Automatically '
                                      'opens a browser tab and shows the '
                                      'file.')
        self.argparser.add_argument('--no-browser', action='store_true',
                                    help=('only creates .html file, but does '
                                          'not open the browser.'))
        self.argparser.add_argument('--out', '-o', action='store',
                                    default=None,
                                    help=('filename to output. Default is '
                                          '<original logfile>.html'))
        self.argparser.add_argument('--line-max', action='store',
                                    default=10000,
                                    help=('max count of datapoints at which '
                                          'actual log line strings are not '
                                          'printed any more.'))

    def _export(self, with_line_str=True):
        fields = ['_id', 'datetime', 'operation', 'thread', 'namespace',
                  'nscanned', 'nreturned', 'duration', 'numYields', 'w', 'r']
        if with_line_str:
            fields.append('line_str')

        first_row = True
        result_str = ''
        out_count = 0
        for line_no, logevent in enumerate(self.args['logfile']):
            # only export lines that have a datetime and duration
            if logevent.duration is not None and logevent.datetime:
                out_count += 1
                # if too many lines include a line_str, the page won't load
                if with_line_str and out_count > self.args['line_max']:
                    print("Warning: more than %i data points detected. "
                          "Skipping actual log line strings for faster "
                          "plotting." % self.args['line_max'])
                    return False
                # write log line out as json
                if not first_row:
                    # prepend comma and newline
                    result_str += ',\n'
                else:
                    first_row = False
                # hack to include _id for log lines from file
                logevent._id = line_no
                result_str += logevent.to_json(fields)
        return result_str

    def run(self, arguments=None):
        LogFileTool.run(self, arguments)

        # store in current local folder
        mlogvis_dir = '.'

        # change stdin logfile name and remove the < >
        logname = self.args['logfile'].name
        if logname == '<stdin>':
            logname = 'stdin'

        if self.args['out'] is not None:
            outputname = self.args['out']
        else:
            outputname = logname + '.html'

        os.chdir(mlogvis_dir)

        data_path = os.path.join(os.path.dirname(mtools.__file__), 'data')
        srcfilelocation = os.path.join(data_path, 'index.html')

        json_docs = self._export(True)
        if not json_docs:
            json_docs = self._export(False)

        outf = ('{"type": "duration", "logfilename": "' +
                logname + '", "data":[' + json_docs + ']}')

        dstfilelocation = os.path.join(os.getcwd(), '%s' % outputname)

        print("copying %s to %s" % (srcfilelocation, dstfilelocation))

        srcfile = open(srcfilelocation)
        contents = srcfile.read()
        srcfile.close()

        dstfile = open(dstfilelocation, 'wt')
        replaced_contents = contents.replace('##REPLACE##', outf)
        dstfile.write(replaced_contents)
        dstfile.close()

        if not self.args['no_browser']:
            print("serving visualization on file://" + dstfilelocation)
            webbrowser.open("file://" + dstfilelocation)


def main():
    tool = MLogVisTool()
    tool.run()


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
"""
mgenerate is no longer included with mtools. Please use
`mgeneratejs <https://www.npmjs.com/package/mgeneratejs>`__ instead.
"""

import sys


def main():
    print("mgenerate is no longer included with mtools. Please use "
          "https://www.npmjs.com/package/mgeneratejs instead.")


if __name__ == '__main__':
    sys.exit(main())



import sys

from mtools.test import all_tools


@all_tools
def test_help(tool_cls):
    """
    Check that all command line tools have a --help option that explains the
    usage.

    As per argparse default, this help text always starts with `usage:`.
    """
    tool = tool_cls()

    try:
        tool.run("--help")

    except SystemExit:
        if not hasattr(sys.stdout, "getvalue"):
            raise Exception('stdout not captured in test.')
        output = sys.stdout.getvalue().strip()
        assert output.startswith('usage:')


@all_tools
def test_version(tool_cls):
    """
    Check that all command line tools have a --version option that returns
    the current version.
    """

    tool = tool_cls()

    try:
        tool.run("--version")

    except SystemExit as ex:
        print(ex)
        if not hasattr(sys.stdout, "getvalue"):
            raise Exception('stdout not captured in test.')

        # argparse's --version outputs to stderr, which can't be captured
        # with nosetests. Therefore just checking that the scripts run and not
        # output anything to stdout --- FIXED to stderr https://bugs.python.org/issue18920
        #output = sys.stdout.getvalue().strip()
        #assert len(output) == 0
